@@ -2,7 +2,7 @@
  * Enhanced Seccomp x86 Specific Code
  *
  * Copyright (c) 2012 Red Hat <pmoore@redhat.com>
- * Author: Paul Moore <pmoore@redhat.com>
+ * Author: Paul Moore <paul@paul-moore.com>
  */
 
 /*
@@ -22,22 +22,8 @@
 #ifndef _ARCH_X86_H
 #define _ARCH_X86_H
 
-#include <stdbool.h>
-
 #include "arch.h"
-#include "db.h"
-#include "system.h"
 
-extern const struct arch_def arch_def_x86;
-
-int x86_syscall_resolve_name(const char *name);
-const char *x86_syscall_resolve_num(int num);
-
-const char *x86_syscall_iterate_name(unsigned int spot);
-
-int x86_syscall_rewrite(const struct arch_def *arch, bool strict, int *syscall);
-
-int x86_filter_rewrite(const struct arch_def *arch, bool strict,
-		       int *syscall, struct db_api_arg *chain);
+ARCH_DECL(x86)
 
 #endif
